from cpython cimport PyObject

cdef extern from "cPoisson.h":
    double log10_poisson_cdf ( unsigned int n, double lam, short lower )
    double log10_poisson_cdf_P_large_lambda ( unsigned int k, double lbd )
    double log10_poisson_cdf_Q_large_lambda ( unsigned int k, double lbd )

